'''
To run FinalA.py, execute the following:
                
        python3 FinalA.py --seed <seed>
'''

import multiprocessing

import pickle
from generator import Generator, Parameters, SurrogateDataset
from reconstruction import Reconstruction
from surrogate import Surrogate
from optimizerA import Optimizer
from matplotlib import pyplot as plt
import time
import torch
from torch.utils.data import DataLoader
import numpy as np
import os
import wandb
import argparse
import random
import json

# Parsing the seed from command line
parser = argparse.ArgumentParser()
parser.add_argument('--seed', type=int, required=True, help='Random seed')
args = parser.parse_args()

# Set the random seed
random.seed(args.seed)

with open('config.json', 'r') as config_file:
    config = json.load(config_file)

def initialise_wandb(run_name = None):
    # Login to wandb using API key
    wandb.login(key= wandb_key)

    wandb.init(
        # set the wandb project where this run will be logged
        project= proj_name,
        name = 'Run' + str(run_num) if run_name is None else run_name,

        # track hyperparameters and run metadata
        config={
            'start_pars' : start_pars,
            'seed' : (args.seed)
        }
    )


def reset_weights(m):
    '''
    Try resetting model weights.
    '''
    for layer in m.children():
        if hasattr(layer, 'reset_parameters'):
            print(f'Reset trainable parameters of layer = {layer}')
            layer.reset_parameters()
            

if __name__ == "__main__":

    multiprocessing.set_start_method('spawn')

# Setting Output Files and WandB run number
    run_num = (args.seed / 100)
    outpath = config["outpath"] + str(run_num) # Adds run_number automatically for instances with multiple runs
    
    os.system('mkdir  '+outpath)
    outpath += '/'
    os.system('cp FinalA.py '+outpath)
    os.system('cp optimizerA.py '+outpath)
    os.system('cp reconstruction.py '+outpath)
    os.system('cp generator.py '+outpath)

    test = False

    divide = 10 if test else 1

    ToP = 2  # Types of Parameters, if we add a new parameter update this by +1 

    start_pars = {
        'thickness_absorber_0': random.uniform(0.1, 50),    # Random float between 0.1 and 50
        'thickness_absorber_1': random.uniform(0.1, 5),
        'thickness_absorber_2': random.uniform(0.1, 50),
        'thickness_scintillator_0': random.uniform(20, 35), # Random float between 0.1 and 50
        'thickness_scintillator_1': random.uniform(0.1, 35),
        'thickness_scintillator_2': random.uniform(0.1, 10),
        'material_abs_0': random.uniform(-1, 1),           # Random float between 0.1 and 1
        'material_abs_1': random.uniform(-1, 1),
        'material_abs_2': random.uniform(-1, 1),
        'material_scint_0': random.uniform(-1, 1),         # Random float between 0.1 and 1
        'material_scint_1': random.uniform(-1, 1),
        'material_scint_2': random.uniform(-1, 1),   
    }

    # Your wandb key and project name here. Incase WandB is not required set below boolean variable to False 
    to_wandb = config['to_wandb']
    wandb_key = config['wandb_key']
    proj_name = config['proj_name']

    initialise_wandb(run_name = config['run_name']) if to_wandb else None  # Run Name is optional arguement, when scheduling multiple runs if run name is not given they automatically are assigned Run0, Run1 ...

    box = np.array(len(start_pars)*[1.5]) #will be adjusted to some extent by optimizer
    #start parameters from run 8
    #start_pars = {'thickness_absorber_0': 0.7642903, 'thickness_absorber_1': 10.469371, 'thickness_scintillator_0': 30.585306, 'thickness_scintillator_1': 22.256506}

    
    gen = Generator(box, 
                    Parameters(ToP, parameters = start_pars),
                    types = ToP,
                    n_vars = 40,
                    n_events_per_var = 400//divide,
                    particles=[['gamma',0.22],['pi+',2.11]])
    

    # reconstruction
    reco_model = Reconstruction(gen.n_parameters, 
                           gen.n_parameters_per_sensor * gen.n_sensors, 
                           gen.n_target_parameters)
    
    surrogate_model = Surrogate(gen.n_parameters, 
                           gen.n_target_parameters + gen.n_context_parameters, #context
                           gen.n_target_parameters,  #reco
                           n_time_steps=50)
    
    optimizer = Optimizer(gen, surrogate_model, reco_model, gen.parameters,
                          constraints= config['constraints']) # length of detector, upper limit, lower limit and max diff of material parameters

    n_epochs_pre = 24//divide
    n_epochs_main = 40//divide
    
    parameters = gen.parameters     
    
    #if saved evolution exists, load it
    import os.path
    if False and os.path.isfile('evolution.npy'):
        evolution = np.load('evolution.npy').tolist()
        print('EVOLUTION:',evolution)
        parameters = evolution[-1]
    
    evolution = []
    def save_evolution(parameters, losses, reco_losses, reco_model, means, std):
        '''
        add to evolution list and save it with pickle.
        save it in human readable format as well
        '''
        ps_dict = gen.translate_parameters(parameters)

        ps = {k: ps_dict[k] for k in ps_dict.keys()}
        ps.update(
            {'reco_model': reco_model.state_dict(),
             'means': means,
             'stds': std}
        )
        evolution.append([ps, losses, reco_losses])

        with open(outpath+'evolution.pkl', 'wb') as f:
            pickle.dump(evolution, f)
        # human readable
        with open(outpath+'evolution.txt', 'w') as f:
            for e in evolution:
                f.write(str(e[0]) + ' ' + str(e[1]) + ' ' + str(e[2]) + '\n')

        # now plot all entries of the dictionary ps in a plot and save it, make a new point for each of the items in the evolution list and use the dictionary keys as legend
        # 'transpose' dictionary
        pltdict = {k: [] for k in ps_dict.keys()}
            
        for i in range(len(evolution)):
            for k in pltdict.keys():
                pltdict[k] += [evolution[i][0][k]]

        # now make a plot with matplotlib
        for k in pltdict.keys():
            style = '-'
            if 'absorber' in k:
                style = '--'
            plt.plot([i for i in range(len(pltdict[k]))], np.where(np.array(pltdict[k]) < 0., 0., np.array(pltdict[k])), label = k, linestyle=style)
            
        plt.legend()
        # create a second y-axis for the loss values. Make the line black and thicker than the others; make the loss y axis log scale
        plt.twinx()
        plt.plot([i for i in range(len(evolution))], [e[1] for e in evolution], label = 'loss', color='black', linewidth=2, alpha=0.2)
        #also plot reco loss also in black but with a different linestyle
        plt.plot([i for i in range(len(evolution))], [e[2] for e in evolution], label = 'reco loss', color='black', linewidth=2, alpha=0.2, linestyle='--')
        plt.yscale('log')
        
        plt.legend()
        plt.xlabel('step')
        plt.savefig(outpath+'evolution_continuous.png')
        wandb.log({"evolution_continuous": wandb.Image(plt)}) if to_wandb else None
        plt.close()

        for k in pltdict.keys():
            if 'material' in k:
                style = '-'
                if 'abs' in k:
                    style = '--'
                plt.plot([i for i in range(len(pltdict[k]))], np.where(np.array(pltdict[k]) < 0., 0., np.array(pltdict[k])*10), label = k, linestyle=style)

        plt.legend()
        plt.xlabel('step')
        plt.savefig(outpath+'evolution_material.png')
        plt.close()

    
    def make_check_plot(surr_out, reco_out, true_in, evolution, outname="" ):
        '''
        make response plots for all three and save them as data in responses.pkl and responses.txt use these files to reproduce the histograms easily. 
        '''
        reco_resp = (reco_out.detach().cpu().numpy() - true_in.detach().cpu().numpy()) / true_in.detach().cpu().numpy()
        surr_resp = (surr_out.detach().cpu().numpy() - true_in.detach().cpu().numpy()) / true_in.detach().cpu().numpy()

        # Un-comment the commented code to save histograms plots directly (clutters the folder)
        # if len(outname) == 0:
        #     outname = str(len(evolution))

        # Compute histograms
        reco_hist, bins = np.histogram(reco_resp.flatten(), bins=51, range=(-1, 1))
        surr_hist, _ = np.histogram(surr_resp.flatten(), bins=51, range=(-1, 1))
    
        # Load existing data if the JSON file exists
        pickle_path = os.path.join(outpath, 'responses.pkl')

        # Load existing data or create empty list
        if os.path.exists(pickle_path):
            with open(pickle_path, 'rb') as f:
                all_histogram_data = pickle.load(f)
        else:
            all_histogram_data = []

        # if(len(evolution)==0):
        #     return
        
        # The data of the response plots. 
        histogram_data = {
            "iteration": len(evolution),
            "reco_hist": reco_hist.tolist(),
            "surr_hist": surr_hist.tolist(),
            "bins": bins.tolist(),
            "evolution_loss": evolution[-1][1] if evolution else None
        }
        all_histogram_data.append(histogram_data)

        # Save the combined histogram data
        with open(pickle_path, 'wb') as f:
            pickle.dump(all_histogram_data, f)

        with open(outpath+'responses.txt', 'w') as f:
            for e in all_histogram_data:
                f.write('{')
                for key, value in e.items():
                    # Format key-value pair into a string and write to file
                    if (key != "evolution_loss"):
                        f.write(f"{key}: {value},")
                    else: 
                        f.write(f"{key}: {value}" + "}")
                f.write('\n')
        # plt.figure(figsize=(10,10))
        # plt.hist(reco_resp.flatten(), bins=51, range=(-1,1), alpha = 0.5, label='G4+reco')
        # plt.hist(surr_resp.flatten(), bins=51, range=(-1,1), alpha = 0.5, label='E2E surrogate')
        # plt.text(0.15, 0.8, 'step '+outname, transform=plt.gca().transAxes)
        # if len(evolution):
        #     plt.text(0.15, 0.7, 'o_loss '+str((evolution[-1][1])), transform=plt.gca().transAxes)
        # plt.legend()
        # plt.savefig(outpath+'response_'+outname+'_lin.png')
        # plt.yscale('log')
        # plt.savefig(outpath+'response_'+outname+'_log.png')
        # plt.close()
    
    def validation(parameters, model_means, model_stds):
        '''
          Generates fresh batch of data and calculates the reconstruction loss for the data. Then puts these results in a dictionary which is returned. 
        '''

        single_ds = gen.create_torch_dataset(parameters, random_scaler=0, means = model_means, stds = model_stds, single=True)

        y_preds, loss = reco_model.apply_model_in_batches(single_ds, 400)

        detector_array = single_ds.detector_array
        context_array = single_ds.context_array
        true_energy = single_ds.target_array
        context_array = context_array.reshape(1,-1)
        true_energy = true_energy.reshape(1,-1)
        y_preds_cpu = y_preds.detach().cpu()

        # # Convert to numpy arrays
        reco_energy = y_preds_cpu.numpy()
        reco_energy = reco_energy.reshape(1, -1)

        iteration_dict = {
            'detector_array': detector_array,
            'context_array': context_array,
            'true_energy': true_energy,
            'reco_energy': reco_energy,
            'loss': loss,
        }

        return iteration_dict
    
    def save_results(iteration_dict, results):
        '''
            Saves and uploads results to results.pkl/results.txt and wandb respectively. 
        '''
        iteration_dict.update({'reco_loss': reco_loss,
                              'o_loss' : o_loss,
                              'reco_surrogate_loss' : reco_surrogate_loss,
                              'constraint_loss': constraint_loss,
                              'scale_factor' : scale_factor })

        # Append the dictionary to the results list and save them 
        results.append(iteration_dict)

        pickle_file = 'results.pkl'
        with open(outpath + pickle_file, 'wb') as f:
            pickle.dump(results, f)

        txt_file = 'results.txt'

        # Open the file in write mode
        with open(outpath + txt_file, 'w') as f:
            for i, result in enumerate(results):
                f.write(f"Iteration {i + 1}:\n")
                f.write(f"Detector Array:\n{result['detector_array']}\n")
                f.write(f"Context Array:\n{result['context_array']}\n")
                f.write(f"True Energy:\n{result['true_energy']}\n")
                f.write(f"Reco Energy:\n{result['reco_energy']}\n")
                f.write(f"Loss:\n{result['loss']}\n")
                f.write(f"Optimizer Loss:\n{result['o_loss']}\n")
                f.write(f"Scale Update:\n{result['scale_factor']}\n")
                f.write("\n")

        print(f"Saved results to {txt_file}")

        # Saving to wandb for mid-run analyses
        wandb_dict = {'loss': iteration_dict['loss'],'o_loss': o_loss,'reco_loss' : reco_loss ,'reco_surrogate_loss' : reco_surrogate_loss,
            'constraint_loss': constraint_loss,  'scale': scale_factor}
        wandb.log(wandb_dict) if to_wandb else None

        parameters_dict = gen.translate_parameters(parameters)
        wandb.log(parameters_dict) if to_wandb else None

        return True

    #pre-train reco and surrogate model
    def pre_train():
        ds = gen.create_torch_dataset(parameters, random_scaler=1.2)
        reco_model.to(config["device"]) #cuda
        print('pre-training 0')
        reco_model.train_model(ds, batch_size=256, n_epochs= 10, lr=0.03)
        print('pre-training 1')
        reco_model.train_model(ds, batch_size=256, n_epochs= n_epochs_pre, lr=0.01)
        print('pre-training 2')
        reco_model.train_model(ds, batch_size=512, n_epochs= n_epochs_pre, lr=0.001)
        print('pre-training 3')
        reco_model.train_model(ds, batch_size=1024, n_epochs= n_epochs_pre, lr=0.001)
        #this is a normalised quantity - at some point only expose the un-normalised ones, but also needs adjustments for surrogate dataset
        reco_result,reco_loss = reco_model.apply_model_in_batches(ds, batch_size=128)

        reco_model.to('cpu')
    
        surrogate_dataset = SurrogateDataset(ds, reco_result.detach().cpu().numpy())#important to detach here
        print('surr pre-training 0')
        surrogate_model.train_model(surrogate_dataset, batch_size=256, n_epochs= 10, lr=0.03)
        print('surr pre-training 1')
        surrogate_model.train_model(surrogate_dataset, batch_size=256, n_epochs= n_epochs_pre, lr=0.01)
        print('surr pre-training 2')
        surrogate_model.train_model(surrogate_dataset, batch_size=512, n_epochs= n_epochs_pre, lr=0.001)
        print('surr pre-training 3')
        surrogate_model.train_model(surrogate_dataset, batch_size=1024, n_epochs= n_epochs_pre, lr=0.001)
        print('surr pre-training 4')
        surrogate_model.train_model(surrogate_dataset, batch_size=1024, n_epochs= n_epochs_pre, lr=0.0003)
        #these are un-normalised quantities
        surr_out, reco_out, true_in = surrogate_model.apply_model_in_batches(surrogate_dataset, batch_size=512)
        make_check_plot(surr_out, reco_out, true_in, evolution, outname = "pretrain")
    print(parameters)
    pre_train()

    best_surrogate_loss = 1e10
    def update_best_surrogate_loss(loss):
        global best_surrogate_loss
        if loss < best_surrogate_loss:
            best_surrogate_loss = loss
            return True
        return loss < 4. * best_surrogate_loss

    covariance_pcas = []
    results = []
    scale_factor = 1  # The Initial Weight of the Sigmoid  
    
    for i in range(1000): #just an example
        
        ds = gen.create_torch_dataset(parameters) # full variations plus a tiny bit
        model_means = ds.c_means
        model_stds = ds.c_stds
        model_means = [tensor.to('cpu').numpy() for tensor in model_means]
        model_stds = [tensor.to('cpu').numpy() for tensor in model_stds]

        reco_model.to(config["device"]) #cuda
        reco_model.train_model(ds, batch_size=256, n_epochs= n_epochs_main//4, lr=0.003)
        reco_model.train_model(ds, batch_size=1024, n_epochs= n_epochs_main//2, lr=0.001)
        reco_model.train_model(ds, batch_size=1024, n_epochs= n_epochs_main//2, lr=0.0003)
        reco_result, reco_loss = reco_model.apply_model_in_batches(ds, batch_size=128)
        reco_model.to('cpu') #free up gpu memory

        #create the surrogate dataset
        surrogate_dataset = SurrogateDataset(ds, reco_result.detach().cpu().numpy())#important to detach here

        print("Surrogate training")
        surrogate_model.train_model(surrogate_dataset, batch_size=1024, n_epochs= n_epochs_main//2, lr=0.005) #get the surrogate model to a good state
        sl = surrogate_model.train_model(surrogate_dataset, batch_size=1024, n_epochs= n_epochs_main, lr=0.0003) #get the surrogate model to a good state
    
        while not update_best_surrogate_loss(sl):
            print("Surrogate retraining")
            pre_train()
            surrogate_model.train_model(surrogate_dataset, batch_size=256, n_epochs= n_epochs_main//5, lr=0.005)
            surrogate_model.train_model(surrogate_dataset, batch_size=1024, n_epochs= n_epochs_main//2, lr=0.005)
            surrogate_model.train_model(surrogate_dataset, batch_size=1024, n_epochs= n_epochs_main//2, lr=0.0003)
            sl = surrogate_model.train_model(surrogate_dataset, batch_size=1024, n_epochs= n_epochs_main//2, lr=0.0001)

        ## check if things worked
        surr_out, reco_out, true_in = surrogate_model.apply_model_in_batches(surrogate_dataset, batch_size=512)
        make_check_plot(surr_out, reco_out, true_in, evolution)

        # Running the reco model on fresh data and storing the results in a dictionary which will be uploaded to WandB
        iteration_dict = validation(parameters, model_means, model_stds)
        
        print("Optimization")
        updated_detector_parameters, optimal, o_loss, reco_surrogate_loss, constraint_loss, scale_factor = optimizer.optimize(surrogate_dataset,ToP, scale_factor, i, batch_size=512, n_epochs=40, lr=0.02, add_constraints=True)#some noise
        print("Updated detector parameters: ", gen.translate_parameters(updated_detector_parameters), "; not interrupted", optimal, "; change", updated_detector_parameters - parameters)

        #use the difference of old and updated parameters to define a direction in parameter space
        #and increase the box size in that direction
        #gen.box_size += updated_detector_parameters - parameters
        # make sure the size does not become too small in any dimension
        #gen.box_size = np.maximum(gen.box_size, 0.1 * np.ones_like(gen.box_size))

        if np.isnan(o_loss) or np.isnan(updated_detector_parameters).any() or np.isinf(o_loss) or np.isinf(updated_detector_parameters).any():
            print("NaN loss, restarting.")
            surrogate_model.apply(reset_weights)
            reco_model.apply(reset_weights)
            pre_train()
            sl = 0. #reset surrogate loss too to avoid infinite loop
            continue

        if save_results(iteration_dict, results) == True: 
            print("Results Saved Succesfully")

        #step more in the direction
        parameters = updated_detector_parameters + 0.5*(updated_detector_parameters - parameters)

        # for bookkeeping, append the generator box_covariance to a text file
        with open(outpath+'box_covariance.txt', 'a') as f:
            f.write(str(gen.box_covariance) + '\n')
        
        # also perform a quick PCA of the box covariance and plot the first two principal components
        # but for all entries of covariance_pcas
        pca = np.linalg.eig(gen.box_covariance)
        covariance_pcas.append(pca)
        plt.plot(range(len(covariance_pcas)), [p[0][0] for p in covariance_pcas])
        plt.plot(range(len(covariance_pcas)), [p[0][1] for p in covariance_pcas])
        plt.xlabel('Step')
        plt.ylabel('1st & 2nd principal component')
        plt.savefig(outpath+f'pca.png')#just overwrite
        plt.close()

        if test:
            print('EVOLUTION:',evolution, gen.box_covariance)
            time.sleep(2)
        save_evolution(updated_detector_parameters, o_loss, reco_loss, reco_model, gen.means, gen.stds) #save the optimised parameters here, not the ones for the next box generation
        #exit()
