import torch
from torch.utils.data import DataLoader
import numpy as np
import json

class Optimizer(object):
    '''
    The optimizer uses the surrogate model to optimise the detector parameters in batches.
    It is also linked to a generator object, to check if the parameters are still in bounds using the function is_local(parameters)
    of the generator.

    Once the parameters are not local anymore, the optimizer will return the last parameters that were local and stop.
    For this purpose, the surrogate model will need to be applied using fixed weights.
    Then the reconstruction model loss will be applied based on the surrogate model output.
    The gradient w.r.t. the detector parameters will be calculated and the parameters will be updated.
    '''
    def __init__(self, generator, surrogate_model, reconstruction_model, 
                 detector_start_parameters,
                 lr=0.001, batch_size=128,
                 constraints : dict = None,):
        with open('config.json', 'r') as config_file:
             config = json.load(config_file)
        self.generator = generator
        self.surrogate_model = surrogate_model
        self.reconstruction_model = reconstruction_model


        self.n_time_steps = surrogate_model.n_time_steps
        self.lr = lr
        self.batch_size = batch_size
        self.constraints = constraints
        self.device = torch.device(config["device"]) #cuda
        self.cu_box = torch.tensor(self.generator.box_size, dtype=torch.float32).to(self.device)
        self.detector_parameters = torch.nn.Parameter(torch.tensor(np.array(detector_start_parameters, dtype='float32')).to(self.device), requires_grad=True)
        
        self.optimizer = torch.optim.Adam([self.detector_parameters], lr=self.lr)
        self.to(self.device)

        
    def to(self, device):
        self.device = device
        self.surrogate_model.to(device)
        self.cu_box.to(device)

    def other_constraints(self, ToP, scale_factor):
        # keep the size of the detector within 3m
        raw_detector_parameters = self.detector_parameters 

        number_parameters = len(self.detector_parameters)

        # Given n parameters, n//ToP will be parameters of each type in order Thickness, Material Choice, and any others added
        divider = int(((number_parameters//ToP)*3)/2) # 3/2 or 1.5 as 1-1.5 will have absorber parameters and 1.5-2 will have material parameters regardless of number of layers.

        thickness_parameters = self.detector_parameters[0:(number_parameters//ToP)] 
        material_parameters = self.detector_parameters[(number_parameters//ToP):(number_parameters//ToP)*2]
        abs_parameters = self.detector_parameters[(number_parameters//ToP): divider ]
        scint_parameters = self.detector_parameters[divider : (number_parameters//ToP)*2]

        # cost_dic = {"G4_POLYSTYRENE": 0.001
        #     ,"G4_PLASTIC_SC_VINYLTOLUENE": 0.001 
        #     ,"G4_BGO": 666.66
        #     ,"G4_LSO": 5000
        #     ,"G4_LYSO": 75000
        #     ,"G4_CESIUM_IODIDE": 0.001
        #     ,"G4_Pb": 25
        #     ,"G4_Fe": 4.166
        #     ,"G4_W": 500
        #     ,"G4_Cu": 66.66
        #     ,"G4_BRASS": 16.66
        #     ,"G4_Si": 8330000
        #     ,"G4_PbWO4": 2500 }
        
        if self.constraints is not None:
            zero_length = (torch.zeros(number_parameters//ToP)).to(self.device)
            negative_length_loss = torch.mean(0.5*torch.nn.ReLU()(zero_length- thickness_parameters)**2)

            if 'length' in self.constraints:
                detector_length = torch.sum(raw_detector_parameters[:(number_parameters//ToP)])# changed this here so as to not include material choice in this
                total_length_loss = torch.mean(10.*torch.nn.ReLU()(detector_length - self.constraints['length'])**2)#constrain it to 25cm

            if 'lower' in self.constraints:
                lbound = (torch.ones(number_parameters//ToP) * self.constraints['lower']).to(self.device)
                lower_loss = torch.mean(0.5*torch.nn.ReLU()(lbound- material_parameters)**2)

            if 'upper' in self.constraints:
                ubound = (torch.ones(number_parameters//ToP) * self.constraints['upper']).to(self.device)
                upper_loss = torch.mean(0.5*torch.nn.ReLU()(material_parameters - ubound)**2)

            if 'diff' in self.constraints:
                abs_maxdiff = max(abs_parameters) - min(abs_parameters)
                scint_maxdiff = max(scint_parameters)- min(scint_parameters)

                diff_loss = torch.mean(100.*torch.nn.ReLU()(abs_maxdiff - self.constraints['diff'])**2)
                diff_loss += torch.mean(100.*torch.nn.ReLU()(scint_maxdiff - self.constraints['diff'])**2)
            if 'cost' in self.constraints: 

                def sigmoid(number):
                    return 1/ (1 + torch.exp(-number))

                abs_sigm = sigmoid(scale_factor*(abs_parameters))  # Scaled sigmoid function
                cost_abs = abs_sigm * 25 + (1 - abs_sigm) * 4.166   # 25 is the cost of Pb and 4.166 for Fe
                scint_sigm = sigmoid(scale_factor*(scint_parameters))
                cost_scint = scint_sigm * 2500 + (1 - scint_sigm) * 0.01 # 2500 is the cost of PbWO4 and 0.01 for Polystyrene. 

                combined = torch.cat((cost_abs, cost_scint), dim=0)

                combined_cost = combined.to(self.device)

                cost = torch.sum(combined_cost * thickness_parameters)
                cost_loss = torch.mean((2/self.constraints['cost'])* torch.nn.ReLU()(cost - self.constraints['cost'])**2)

        # now keep parameters such that within the box size of the generator, there are always some positive values even if the 
        # central parameters are negative. Both box size and raw_detector_parameters are in non-normalised space, so this is straight forward
        # the generator will have to provide the box size
        # this will avoid mode collapse

        # box = self.cu_box # the box has same ordering as raw_detector_parameters
        # lower_para_bound = -box/1.1
        # bloss = torch.mean(100.*torch.nn.ReLU()(lower_para_bound - raw_detector_parameters)**2)
        # + bloss 

        return total_length_loss +lower_loss + upper_loss + cost_loss + negative_length_loss
    
    def clamp_parameters(self):
        return
        self.detector_parameters.data = self.detector_parameters.data.clamp(1e-3) #DEBUG, NE

    def adjust_generator_covariance(self, direction, min_scale=2.0):
        #stretches the box_covariance of the generator in the directon specified as input
        #direction is a vector in parameter space
        v = direction
        v_length = np.linalg.norm(v)
        epsilon = 1e-4
        v_norm = v / (v_length + epsilon)

        s = min_scale *np.max([1., 4.*v_length])  # scale factor at least by a factor of two, if not more

        # Create the scaling adjustment matrix
        M_scaled = (s - 1) * np.outer(v_norm, v_norm)        

        # Adjust the original covariance matrix
        self.generator.box_covariance = np.diag(self.generator.box_size**2) + M_scaled
        # print('new box_covariance', self.generator.box_covariance)


    def optimize(self, dataset, ToP , scale_factor, iteration ,batch_size, n_epochs, lr, add_constraints = False):
        '''
        keep both models fixed, train only the detector parameters (self.detector_start_parameters)
        using the reconstruction model loss
        '''
        # set the optimizer
        for param_group in self.optimizer.param_groups:
            param_group['lr'] = lr

        self.surrogate_model.eval()
        self.reconstruction_model.eval()

        # create a dataloader for the dataset, this is the surrogate dataloader
        data_loader = DataLoader(dataset, batch_size=batch_size, shuffle=True)

        # save the initial parameters
        initial_parameters = np.copy(self.detector_parameters.detach().cpu().numpy())

        # loop over the batches
        mean_loss = 0
        reco_surrogate_loss = 0
        constraint_loss = 0 
        for epoch in range(n_epochs):
            batch_mean_loss = 0 # To use last epochs only
            batch_reco_surrogate_loss = 0
            batch_constraint_loss = 0 
            stop_epoch = False
            for batch_idx, (_, true_inputs, true_context, reco_result) in enumerate(data_loader):
                # in principle this could also be sampled from the correct distributions; but the distributions are not known in all cases (mostly for context)
                # keep in mind for an extension
                true_inputs = true_inputs.to(self.device)
                true_context = true_context.to(self.device)
                reco_result = reco_result.to(self.device)
                # apply the model
                n_detector_paras  = dataset.normalise_detector(self.detector_parameters)
                reco_surrogate = self.surrogate_model.sample_forward(n_detector_paras, 
                                                                     true_inputs, 
                                                                     true_context)
                # calculate the loss
                loss =  self.reconstruction_model.loss(dataset.unnormalise_target(reco_surrogate), dataset.unnormalise_target(true_inputs))
                batch_reco_surrogate_loss += loss.item()
                if add_constraints:
                    constraint_loss_value = self.other_constraints(ToP, scale_factor)
                    batch_constraint_loss += constraint_loss_value.item()
                    loss += constraint_loss_value
                
                self.optimizer.zero_grad()
                loss.backward()
                #print('gradient',self.detector_parameters.grad, 'should have', self.detector_parameters.requires_grad)

                #if loss is nan, stop
                if np.isnan(loss.item()):
                    print("NaN loss, exiting.")
                    # save parameters, reset the optimiser as if it made a step but without updating the parameters
                    prev_parameters = self.detector_parameters.detach().cpu().numpy()
                    self.optimizer.step()
                    self.detector_parameters.data = torch.tensor(prev_parameters).to(self.device)
                    # return 
                    return self.detector_parameters.detach().cpu().numpy(), False, mean_loss / (batch_idx+1)
                
                self.optimizer.step()
                batch_mean_loss += loss.item()

                #record steps

                #check if the parameters are still local otherwise stop
                if not self.generator.is_local(self.detector_parameters.detach().cpu().numpy(),0.8):#a bit smaller box size to be safe
                    stop_epoch = True
                    stopped_at = epoch
                    print("parameter not local")
                    break
                
                if batch_idx % 20 == 0: 
                    nppars = self.detector_parameters.detach().cpu().numpy()
                    pdct = self.generator.translate_parameters(nppars)
                    print('current parameters: ')
                    for k in pdct.keys():
                        print(k, pdct[k])
            batch_mean_loss /= (batch_idx+1)
            batch_reco_surrogate_loss/= (batch_idx+1)
            batch_constraint_loss/= (batch_idx+1)
            if epoch == (n_epochs - 1): # Only using Last Epochs(Remove this conditional to use more)
                mean_loss += batch_mean_loss
                reco_surrogate_loss += batch_reco_surrogate_loss
                constraint_loss+= batch_constraint_loss 
            print('Optimizer Epoch: {} \tLoss: {:.8f}'.format(
                        epoch, loss.item()))
            # stopped_at = epoch+1
            if stop_epoch:
                break
        if not stop_epoch:
            scale_factor += 0.08 # Simplest approach that works well
            print('Scale updated to : ', scale_factor)
        # scale_update += 1
        self.clamp_parameters()
        # commented code useful to calculate loss average instead of loss of only last epoch
        # constraint_loss/=  stopped_at 
        self.adjust_generator_covariance( self.detector_parameters.detach().cpu().numpy() - initial_parameters )
        return self.detector_parameters.detach().cpu().numpy(), True, mean_loss, reco_surrogate_loss, constraint_loss ,scale_factor

    def get_optimum(self):
        return self.detector_parameters.detach().cpu().numpy()









    

