import matplotlib.pyplot as plt
import matplotlib.animation as animation
import numpy as np

def smoothen_dependence(orig_array_xyz, oversampling=1):
    # smoothed_data has the form (n, c, m), where n is the number of frames
    # c is the number of channels in the data, e.g. x, y, z, or r, g, b, or
    # and m is the number of data points in each frame
    # we want to interpolate between each frame to get a smoother
    # animation
    # oversampling is the number of frames to be interpolated between each
    # original frame the interpolation should use a numpy function and should not be done via a loop
    n, c, m = orig_array_xyz.shape
    new_array_xyz = np.zeros((n * oversampling, c, m))
    for i in range(c):
        for j in range(m):
            new_array_xyz[:, i, j] = np.interp(np.arange(n*oversampling)/ oversampling, np.arange(n), orig_array_xyz[:, i, j])

    return new_array_xyz


def update_plot(frame_number, ax, smoothed_data):
    # smoothed_data has the form (n, c, m), where n is the number of frames
    # c is the number of channels in the data, e.g. x, y, z, or r, g, b, or
    # and m is the number of data points in each frame
    # plot is an axis object or a pyplot object
    # frame_number is the frame number to be plotted
    # y_index is the index of the channel to be plotted
    # get x_points the axis x points that are constant during animation
    x_points = ax.lines[0].get_xdata()
    for y_index in range(smoothed_data.shape[1]):
        ax.lines[y_index].set_data(x_points, smoothed_data[frame_number, y_index])
    return ax

def animate_smoothed_data(smoothed_data, 
                          labels,
                          x_points,
                          oversampling=20,
                          x_label="",
                          y_label="",
                          interval=50,
                          outfile="smoothed_data.gif"):
    # smoothed_data has the form (n, c, m), where n is the number of frames
    # x_points is an array describing the x axis points that are constant during animation
    # c is the number of channels in the data, e.g. x, y, z, or r, g, b, or
    # and m is the number of data points in each frame
    # labels is a list of strings with the names of the channels
    # oversampling is the number of frames to be interpolated between each
    # original frame
    # interval is the time in milliseconds between each frame

    # check if the label list length matches the provided number of channels to be plotted
    if len(labels) != smoothed_data.shape[1]:
        raise ValueError('The number of labels does not match the number of channels in the data')
    #check if the x points match the number of data points in each frame
    if len(x_points) != smoothed_data.shape[2]:
        raise ValueError('The number of x points does not match the number of data points in each frame')

    smoothed_data = smoothen_dependence(smoothed_data, oversampling=oversampling)
    fig, ax = plt.subplots()
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.set_xlim(x_points[0], x_points[-1])

    #set y limits to maximum and minimum of the smoothed data
    ax.set_ylim(np.min(smoothed_data), np.max(smoothed_data))
    
    # set the x ticks and labels, but only if the x_points are not too many
    if len(x_points) < 20:
        ax.set_xticks(x_points)
        ax.set_xticklabels([str(x) for x in x_points])
    # otherwise just set the number of ticks to 10
    else:
        ax.set_xticks(np.linspace(x_points[0], x_points[-1], 10))

    ax.grid()
    #ax.set_title('Animation of smoothed data')
    lines = [ax.plot(x_points, smoothed_data[0, i], label=labels[i])[0] for i in range(smoothed_data.shape[1])]
    ax.legend()
    # make an animation for all channels in the data
    ani = animation.FuncAnimation(fig, update_plot, frames=smoothed_data.shape[0], 
                                  fargs=(ax, smoothed_data), interval=interval//oversampling + 1, blit=False)
    # show the animation

    ani.save(outfile, fps=30, writer='ffmpeg', )
    #plt.show()

    
def test():
    # create some test data
    n = 100
    m = 100
    x = np.linspace(0, 10, m)
    y = np.sin(x)
    z = np.cos(x)
    data = np.zeros((n, 3, m))
    for i in range(n):
        data[i, 0] = np.sin(x + i / 10)
        data[i, 1] = np.cos(x + i / 10)
        data[i, 2] = np.tan(x + i / 10)
    
    # create some labels
    labels = ['sin(x)', 'cos(x)', 'tan(x)']
    
    # animate the data
    animate_smoothed_data(data, labels, x, x_label='x', y_label='y', interval=50, outfile='test.gif')

